//#define GCC_COMPILER

#include  <msp430x20x3.h>
#include <stdlib.h>
#include <stdio.h>

#ifdef GCC_COMPILER
#include <signal.h>
#endif


/************************************************************************
MACROS
***********************************************************************/ 
#ifndef true
#define true 1
#endif
#ifndef TRUE
#define TRUE 1
#endif
#ifndef false
#define false 0
#endif
#ifndef FALSE
#define FALSE 0
#endif

#define CONV_MODE_SINGLE 1
#define CONV_MODE_CONTINOUS 2

//commands:
#define SLAVE_SET_NR_CONV 1
#define SLAVE_SET_RATE 2
#define SLAVE_SET_CONTINOUS 4
#define SLAVE_SET_SINGLE 8
#define SLAVE_SET_RESET 16

#define DEFAULT_RATE 0.05098
#define DEFAULT_SAMPLES_COLLECT 255;

/************************************************************************
Globals
***********************************************************************/ 
unsigned char SLV_Addr = 0x00;      // Our I2C address
int I2C_State = 0;         // State variable
char  timer_trigger = false;  // turns true every fifty millisecond hast to be cleared
char  adc_reading_ready = false;  // turns true whenever new conversion results are available
unsigned int sda16_result = 32768;
//unsigned long system_time = 0; // keeps sytem time in 65536 microsecond increments

unsigned char in_buffer[3]; // buffer for incoming I2C data
unsigned char out_buffer[3];// buffer for outgoing I2C data

unsigned char I2C_byte_counter = 0; //counting number of bytes received and/or sent

float adc_buffer;  // cummulative moving average filter buffer
unsigned int adc_buffer_int;// temporary conversion variable for cummulative filter buffer
float mavg_rate = DEFAULT_RATE;    // inverse of moving average learninng rate
unsigned char number_of_samples = 0; //number of samples in the buffer
unsigned char samples_to_collect = DEFAULT_SAMPLES_COLLECT; //number of samples to collect in single mode
unsigned char conversion_mode = CONV_MODE_CONTINOUS; //conversion mode either CONV_MODE_SINGLE or *_CONTINOUS

#pragma location=0xbf  //to fix the SDA2 errata (wrong temperature coefficient)
__no_init char errata_SDA2;


/************************************************************************
Main

Originally the initialization for different periferals was in their own
functions, but I had to start moving them in here to save on the 
memory footprint

***********************************************************************/ 
int main(void)
{ 
	char i=60;  //counter to slow down LED flashing

	//+++++++++++++++++++++++++++++++++++++++++++
	//init watchdog  -  deactivate watchdog timer
	WDTCTL = WDTPW + WDTHOLD; 

	//+++++++++++++++++++++++++++++++++++++++++++
	//fix temperature coefficient of SDA-16  (see errata SDA2)
	errata_SDA2 = 0x61;

	//+++++++++++++++++++++++++++++++++++++++++++
	//init_clock();
	BCSCTL1 = 128+8+4+2+1;   // Set DCO to more or less 20MHz turn off XT2
	DCOCTL = 128+64+32;
	BCSCTL2 = SELM_0; //DCOCLK is sourcing MCLK

	//+++++++++++++++++++++++++++++++++++++++++++  
	//init_timer();  //we use the timer for flashing the green LEDs
	TACTL = TACLR; //=4 resets TACLEAR BIT
	TACTL = TASSEL_2 | MC_2 | ID_2 | TAIE; //SMCLK


	//+++++++++++++++++++++++++++++++++++++++++++  
	//init_ports();
	P1DIR = 0x08;  //Vref_Out    
	P1SEL = 0x08;  //Vref_Out 
	P1IE = 0x10; //enable interrupt on P1.4 (triggering)
	P1IES = 0; //low to high transition causes interrputs (on all inputs :)
	P1IFG = 0;  // clears all port 1 IFGs 
	P2SEL = 0; //P2 is all ports not periferals
	P2DIR = 192; //P2 is all outputs
	P2OUT = 192; //turn off LEDs  

	//+++++++++++++++++++++++++++++++++++++++++++  
	//  init_USI();
	USICTL0 = USISWRST;       // Temporarily turn off USI
	USICTL0 |= USIPE6+USIPE7; // Enable SDA and SCL ports, MSB first, Slave mode
	USICTL1 = USII2C+USIIE+USISTTIE; // Enable I2C mode & USI interrupts (start and counter)
	USICKCTL = USICKPL;             // Setup clock polarity (inactive high)z
	USICNT |= USIIFGCC;             // Disable automatic clear of USIIFG
	USICTL0 &= ~USISWRST;           // Enable USI
	USICTL1 &= ~USIIFG;             // Clear pending flag

	//+++++++++++++++++++++++++++++++++++++++++++  
	//init_SD16();
	//SD16CTL = SD16REFON + SD16SSEL_1;   //1.2V ref, SMCLK
	SD16CTL = SD16REFON + SD16SSEL_1 + SD16XDIV_2 + SD16DIV_1;  //1.2V ref, SMCLK 52KHz sampling clock at 20MHz
	SD16INCTL0 = SD16INCH_0 + SD16GAIN_32;  //select A0  and set the gain to 32
	//  SD16INCTL0 = SD16INCH_0 + SD16GAIN_1;  //select A0  and set the gain to 1 
	SD16CCTL0 =  SD16OSR_1024 + SD16IE;     //2034 OSR, bipolar, interrupt enable
	//SD16CCTL0 =  SD16OSR_1024 + SD16IE + SD16LSBACC;     //2034 OSR, bipolar, interrupt enable, LSB access
	SD16AE = SD16AE0 +SD16AE1;              // P1.0 A0+, A0-

#ifndef GCC_COMPILER
	_EINT(); //enable interrupts
#else
	eint();
#endif

	while(1)
	{    
		//take care of Timer events
		if(timer_trigger)
		{
			timer_trigger=false;
			i--;
			if(!i) //if i counted all the way down
			{
				if(SLV_Addr) //if slave address is set - quick flashing
					i = 20;
				else  //otherwise slow flashing
					i = 60;
				P2OUT ^= 0x80;  //change P2.7 (xor)
			}
		}

		//make sure ADC is running
		if((conversion_mode==CONV_MODE_CONTINOUS) ||
			(number_of_samples<samples_to_collect))
		{
			if(!(SD16CCTL0 & SD16SC))
				SD16CCTL0 |= SD16SC; //start off ADC16
		}

		//if there is a new reading available - deal with it
		if(adc_reading_ready)
		{
                        adc_reading_ready = false;
			if(!number_of_samples)
			{
				adc_buffer=sda16_result;

				//we need to disable interrupts so we can modify
				//adc buffer without the chance that during this
				//operation someone requests its data (happens!)
#ifndef GCC_COMPILER
				_DINT(); //enable interrupts
#else
				dint();
#endif
				adc_buffer_int = (unsigned int)adc_buffer;
				number_of_samples++;               
#ifndef GCC_COMPILER
				_EINT(); //enable interrupts
#else
				eint();
#endif
			}
			else if((number_of_samples<samples_to_collect)||
				(conversion_mode==CONV_MODE_CONTINOUS))
			{
				adc_buffer *= (1.0-mavg_rate);
				adc_buffer += mavg_rate*sda16_result;

				//we need to disable interrupts so we can modify
				//adc buffer without the chance that during this
				//operation someone requests its data (happens!)
#ifndef GCC_COMPILER
				_DINT(); //enable interrupts
#else
				dint();
#endif
				adc_buffer_int = (unsigned int)adc_buffer;
				if(number_of_samples<255)
					number_of_samples++;  
#ifndef GCC_COMPILER
				_EINT(); //enable interrupts
#else
				eint();
#endif	      
				P2OUT ^= 0x40; //flash yellow LED
			}
			else 
			{
				SD16CCTL0 &= ~SD16SC; //turn off ADC16
				P2OUT |= 0x40; //turn of yellow LED
			}
		}
	}
}

//******************************************************************************
// USI interrupt service routine
//******************************************************************************
#ifndef GCC_COMPILER
#pragma vector = USI_VECTOR
__interrupt void USI_TXRX (void)
#else
interrupt (USI_VECTOR) USI_TXRX(void)
#endif
{
	if (USICTL1 & USISTTIFG)   // Start condition received
	{
		I2C_State = 2;         // Enter 1st state on start
	}

	switch(I2C_State)
	{
		//case 0: // Idle, receiving while out of state
		//  break;

	case 2: // we get here at start condition
		USICNT = (USICNT & 0xE0) + 0x08; // Bit counter = 8, RX address and RW
		USICTL1 &= ~USISTTIFG;   // Clear start flag
		I2C_State = 4;           // Go to next state: check address
		break;

	case 4: // We have now received 8 bits of address and R/W
		if (USISRL & 0x01)       // If read (master wants data from us)
		{                      //(last bit received is a 1)
			USICTL0 |= USIOE;    // enable SDA output as we need to reply
			if (((USISRL&254) == (SLV_Addr<<1))  
				&&(SLV_Addr))  // Address match (and is set)
			{  //this means that the master is asking us for results. assemble results.
				USISRL = 0x00;         // Send Ack
				//P1OUT ^= 0x10;  //change P1.4 (xor) 
				//assemble report:
				I2C_byte_counter=0;
				out_buffer[0] = adc_buffer_int & 0xFF; //low byte
				out_buffer[1] = (adc_buffer_int>>8) & 0xFF; //high byte
				out_buffer[2] = number_of_samples;
				I2C_State = 12;         // Go to next state: TX data
			}
			else
			{
				USISRL = 0xFF;         // Send NAck
				I2C_State = 6;         // Go to next state: prep for next Start
			}
		}
		else  //write (receive data from master)
		{
			USICTL0 |= USIOE;    // enable SDA output as we need to reply
			if((!USISRL) && (P1IN & 0x04)) //we are going to receive our address
			{
				//P1OUT ^= 0x08;  //change P1.3 (xor)   
				I2C_byte_counter=0;
				USISRL = 0x00;         // Send Ack
				I2C_State = 18;         // Go to next state: RX data but this is address mode
			}
			else if ((USISRL&254) == (SLV_Addr<<1))  // Address match?
			{
				I2C_byte_counter=0;
				USISRL = 0x00;         // Send Ack
				I2C_State = 8;         // Go to next state: RX data, command mode
			}
			else
			{
				USISRL = 0xFF;         // Send NAck
				//P1OUT |= 0x20;         // LED of: error
				I2C_State = 6;         // Go to next state: prep for next Start
			}
		}
		USICNT |= 0x01;  // send one bit  
		break;

	case 6: // Prep for Start condition (we were not the destination)
		USICTL0 &= ~USIOE;   // SDA = input
		I2C_State = 0;       // Reset state machine
		break;

	case 8: // Receive data byte (command mode)
	case 18: // Receive data byte (address_mode)
		USICTL0 &= ~USIOE;       // SDA = input
		USICNT |=  0x08;         // Bit counter = 8, RX 8 bits of data
		I2C_State+=2;          // Go to next state: Test data and (N)Ack
		break;

	case 10:// Check Data & TX (N)Ack command mode
		USICTL0 |= USIOE;        // SDA = output
		in_buffer[I2C_byte_counter]=USISRL;
		I2C_byte_counter++;
		if(I2C_byte_counter==3) //last byte has received
		{
			switch(in_buffer[0])
			{
			case SLAVE_SET_NR_CONV: //we got a set conversion command
				number_of_samples=0;
				samples_to_collect=in_buffer[1];
				break;
			case SLAVE_SET_RATE: //we got a set rate command
				number_of_samples=0;
				mavg_rate=in_buffer[1];
				mavg_rate*=0.003921568;  //inverse of 255
				break;
			case SLAVE_SET_CONTINOUS: //we got a set continous command
				conversion_mode=CONV_MODE_CONTINOUS;
				break;
			case SLAVE_SET_SINGLE: //we got a set single command
				number_of_samples=0;
				conversion_mode=CONV_MODE_SINGLE;
				break;
			case SLAVE_SET_RESET: //we got a soft reset command
				number_of_samples=0;
				break;
			}
			I2C_State = 6;  // Go to next state: prep for next Start
		}
		else
		{
			I2C_State = 8;  // Go to next state: prep for next Start
		}
		USISRL = 0x00;         // Send Ack
		USICNT |= 0x01;          // Bit counter = 1, send (N)Ack bit
		break;

	case 20:// Check Data & TX (N)Ack address mode
		USICTL0 |= USIOE;        // SDA = output
		SLV_Addr = USISRL & 127;
		USISRL = 0x00;         // Send Ack
		USICNT |= 0x01;          // Bit counter = 1, send (N)Ack bit
		I2C_State = 6;           // Go to next state: prep for next Start
		break;

	case 12: // Transmit data byte (we are the one asked)
		USICTL0 |= USIOE;       // SDA = output
		USISRL = out_buffer[I2C_byte_counter];
		I2C_byte_counter++;
		USICNT |=  0x08;         // Bit counter = 8, RX 8 bits of data
		I2C_State = 14;          // Go to next state: Test data and (N)Ack
		break;

	case 14: // receive ACK
		USICTL0 &= ~USIOE;       // SDA = input
		USICNT |=  0x01;         // Bit counter = 8, RX 8 bits of data
		if(I2C_byte_counter==3)
			I2C_State = 6;          // Go to next state: Test data and (N)Ack
		else
			I2C_State = 12;
		break;
	}

	USICTL1 &= ~USIIFG;                  // Clear pending flags
}


/************************************************************************
SYSTEM TIMER A Interrupt Handlers
***********************************************************************/ 
/*#ifndef GCC_COMPILER
#pragma vector=TIMERA0_VECTOR
__interrupt void Timer_A0 (void)
#else
interrupt (TIMERA0_VECTOR) Timer_A0(void)
#endif
{
//  for generating a 50ms interrupt
TACCR0 += 50000;
timer_trigger = true;
}
*/

#ifndef GCC_COMPILER
#pragma vector=TIMERA1_VECTOR
__interrupt void Timer_A1 (void)
#else
interrupt (TIMERA1_VECTOR) Timer_A1(void)
#endif
{
	//this interrupt hits when timerA counter overflows, thus every 0x10000 
	//units (unit is 8/20Mhz) ; Timer A is used to enable flashing (coarse timing)
	timer_trigger = true;
	TACTL &= ~TAIFG; 
}

/************************************************************************
ADC16 (SDA16) Interrupt Handler
***********************************************************************/ 
#ifndef GCC_COMPILER
#pragma vector = SD16_VECTOR
__interrupt void SD16ISR(void)
#else
interrupt (SD16_VECTOR) SD16ISR(void)
#endif
{
	sda16_result=SD16MEM0;
	adc_reading_ready = true;
}

/************************************************************************
Port 1 Interrupt Handler
***********************************************************************/ 
#ifndef GCC_COMPILER
#pragma vector = PORT1_VECTOR
__interrupt void Port1ISR(void)
#else
interrupt (PORT1_VECTOR) Port1ISR(void)
#endif
{
	SD16CCTL0 &= ~SD16SC; //stop ADC16
	P1IFG = 0; //clear all port 1 IFGs
	//P1IFG &= ~0x10;  // P1.4 IFG cleared
	//P1OUT ^= 0x08;  //change P1.3 (xor)   
	number_of_samples=0;
	SD16CCTL0 |= SD16SC; //start off ADC16
}
